"""
Parse the definition document:

    http://genome.ucsc.edu/goldenPath/help/trackDb/trackDbLibrary.shtml

into various useful forms.  The end goal is to automatically construct
validation (in trackhub.constants), but this is a long-term goal. For now, we
just want to make sure all parameters are represented.


In addition, this script converts the html and various div tag info into ReST
format, for eventual inclusion in the trackhub docs.


The definition document contains "settings blurbs" in <div> tags with
a specific format.  Here's a cheat-sheet:


    <div class="longLabel">
    -----------------------
    - class name usually reflects the param name
    - "*_example" will be ignored for now

    <span class="types all"</span>
    ------------------------------
    - span.attrs['class'][1:] are the track types the setting is valid for

    <div class="format">
    --------------------
        - settings DIVs must contain a sub-div with class "format"

        <code>longLabel</code></div>
        ----------------------------
            - the actual name of the setting is here, as the only content of
              the "format" div -- and is wrapped in code tags

    multiple <p> tags
    -----------------
        - content and description (converted to ReST)

"""
import subprocess
from bs4 import BeautifulSoup
import re
from trackhub import constants


def format_handler(s):
    """
    Convert a format string into a set() to [eventually] be used for validation

    TODO: still needs lots of work
    """
    int_range = re.compile('\d+-\d+')

    s = s.replace('>', '').replace('<', '')

    defined_options = s.split('/')

    option_set = set()

    remainder = []

    # detect int ranges
    for op in defined_options:
        m = int_range.search(op)
        if m:
            substr = op[m.start():m.end()]
            start, stop = substr.split('-')
            option_set.update(range(int(start), int(stop) + 1))
            remaining = s.replace(substr, '').replace('<', '').replace('>', '')
        else:
            remainder.append(op)
    option_set.update(remainder)

    return option_set


def html2rst(html):
    """
    Requires pandoc (sudo apt-get install pandoc); converts HTML string to ReST
    string
    """
    # thanks to
    # http://johnpaulett.com/2009/10/15/ \
    #        html-to-restructured-text-in-python-using-pandoc/
    p = subprocess.Popen(
        [
            'pandoc',
            '--from=html',
            '--to=rst',
            '--strict',
        ],
        stdin=subprocess.PIPE, stdout=subprocess.PIPE)
    return p.communicate(html)[0]


soup = BeautifulSoup(open('trackDbLibrary.shtml').read())

# Only consider settings for these track types
hub_types = set(['bam', 'bigBed', 'bigWig', 'vcfTabix', 'all'])
supported_params = {}
for div in soup.find_all('div'):

    # If we don't have this span class, it's not a settings blurb
    if not div.find_all('span', class_='types'):
        continue

    tag = div['class'][0]

    # example divs aren't useful for now
    if '_example' in tag:
        continue

    # As per the comments in the definition document, settings blurb divs are
    # required to have a sub-div of class "format"
    assert div.div['class'][0] == 'format'

    # Only hub-supported settings for now
    if 'NOT FOR HUBS' in div.text:
        if 'Not yet supported by bigBeds' not in div.text:
            continue

    types = div.span.attrs['class'][1:]

    # Skip if the setting doesn't apply to track hub types
    if len(set(types).intersection(hub_types)) == 0:
        continue

    # TODO: how to handle "sub-settings"?  E.g., url; bed; scoreFilter, etc
    definition = div.div.text.strip().replace('\n', '    ')

    content = html2rst(str(div))

    block = []
    block.append(tag)
    block.append('=' * len(tag))
    block.append(':definition: ``%s``' % definition)
    block.append('')
    block.append(':used for: ``%s``' % types)
    block.append('')

    items = definition.split()
    key = items[0]
    if len(items) == 1:
        format_string = ""
    else:
        format_string = items[1]

    block.append(":options: ``%s``" % format_handler(format_string))
    block.append('')
    block.append(content)

    supported_params[key] = '\n'.join(block)


existing_params = set(
    constants.track_fields.keys()
    + constants.track_typespecific_fields.keys()
    + constants.view_track_fields.keys()
    + constants.composite_track_fields.keys()
    + constants.aggregate_track_fields.keys()
)

print "Add these params to trackhub.constants:"
print "---------------------------------------"
for k, v in supported_params.items():
    if k not in existing_params:
        print v



import os
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
from validate import ValidationError
from base import HubComponent


class Hub(HubComponent):
    # map proper track hub stanza field names to pythonic attribute names in
    # this class.
    _field_order = OrderedDict([
        ('hub', 'hub'),
        ('shortLabel', 'short_label'),
        ('longLabel', 'long_label'),
        ('genomesFile', 'genomes_filename'),
        ('email', 'email'),
    ]
    )

    def __init__(self, hub, short_label="", long_label=None,
                 genomes_file=None, genomes_filename=None, email="",
                 url=None):
        """
        Represents a top-level track hub container.
        """
        HubComponent.__init__(self)
        self.url = url
        self._local_fn = None
        self._remote_fn = None
        self._local_dir = None
        self._remote_dir = None
        self.hub = hub
        self.short_label = short_label
        if not long_label:
            long_label = short_label
        self.long_label = long_label
        self.email = email

        self.genomes_file = None
        if genomes_file is not None:
            self.add_genomes_file(genomes_file)

    @property
    def local_fn(self):
        if self._local_fn is not None:
            return self._local_fn
        return os.path.join(self.local_dir, self.hub + '.hub.txt')

    @local_fn.setter
    def local_fn(self, fn):
        self._local_fn = fn

    @property
    def local_dir(self):
        if self._local_dir is not None:
            return self._local_dir
        return ""

    @local_dir.setter
    def local_dir(self, fn):
        self._local_dir = fn

    @property
    def remote_fn(self):
        if self._remote_fn is not None:
            return self._remote_fn
        return os.path.join(self.remote_dir, self.hub + '.hub.txt')

    @remote_fn.setter
    def remote_fn(self, fn):
        self._remote_fn = fn

    @property
    def remote_dir(self):
        if self._remote_dir is not None:
            return self._remote_dir
        return ""

    @remote_dir.setter
    def remote_dir(self, fn):
        self._remote_dir = fn

    def validate(self):
        if self.genomes_file is None:
            raise ValidationError(
                'No GenomesFile attached to Hub (use add_genomes_file())')
        self.genomes_file.validate()
        return True

    def add_genomes_file(self, genomes_file):
        """
        If a GenomesFile object was not provided upon instantiating this
        object, attach one now
        """
        self.genomes_file = self.add_child(genomes_file)

    def __str__(self):
        s = []
        for field, attr in self._field_order.items():
            if field == 'genomesFile':
                if self.genomes_file:
                    value = os.path.relpath(
                        self.genomes_file.local_fn,
                        start=os.path.dirname(self.local_fn))
                else:
                    value = None
            else:
                value = getattr(self, attr)
            s.append('%s %s' % (field, value))
        return '\n'.join(s)

    def _render(self):
        """
        Render just this object, and not all the underlying GenomeFiles and
        their TrackDb.
        """
        fout = open(self.local_fn, 'w')
        fout.write(str(self))
        fout.close()
        return fout.name



try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
from validate import Parameter
import validate


# http://genome-source.cse.ucsc.edu/gitweb/
#       ?p=kent.git;a=blob;f=src/hg/makeDb/trackDb/README;hb=HEAD


track_field_order = OrderedDict((i.param, i) for i in [
    Parameter(
        'track',
        "track name",
        str),
    Parameter(
        'bigDataUrl',
        "URL",
        str),
    Parameter(
        'shortLabel',
        "label used on left-side labels",
        str),
    Parameter(
        'longLabel',
        'label used in center',
        str),
    Parameter(
        'type',
        'track type',
        # TODO: better validation here for things like bed 3
        str),
])

track_typespecific_fields = {
    None: (),

    'bigBed': OrderedDict((i.param, i) for i in [
        Parameter(
            'colorByStrand',
            """
            red, green, blue  red, green, blue values 0 to 255;
            specifies plus and minus strand color, first rgb is plus
            strand, second rgb is minus strand
            """,
            # NOTE: BED 6 or higher
            validate.RGBList),

        Parameter(
            'denseCoverage',
            """
            in dense mode, do a density plot based on max coverage seen under
            pixel. The maxVal corresponds to the count at which it gets as dark
            as it can get. If maxVal is 0 then this will be calculated from the
            data itself.""",
            int),

        Parameter(
            'exonArrow',
            """
            on|off; shows arrows on blocks allowing user to navigate to next
            block for an item outside the image currently displayed (Default:
            on)
            """,
            # NOTE: BED 10 or higher
            set(['on', 'off'])),

        Parameter(
            'scoreFilter',
            """default score filter value for a track which excludes scores
            below threshold""",
            # TODO: Verify documentation confusion, possibly could be using
            # range
            int),

        Parameter(
            'scoreFilterLimits',
            """
            low[:high]; range that score is filtered between, score value
            N imples N:1000 (Default 0:1000)
            """,
            str),

        Parameter(
            'itemRgb',
            """
            on|off; activates item coloring based on ninth field (Default: off)
            """,
            # NOTE: BED 9 or higher
            set(['on', 'off'])),

        Parameter(
            'maxItems',
            """
            integer defining threshold of items to display in full mode and
            number of lines to display in pack mode. (Default: 250, can't be
            larger than 100,000)
            """,
            # TODO: Improve verification to check that values is less than
            # 100,000
            int),

        Parameter(
            'minGrayLevel',
            """
            1-9; specifies the lightest shade to be used
            """,
            # TODO: Need to look up better verification for range
            int),

        Parameter(
            'noScoreFilter',
            """
            on|off; turns off filter on score configuration options in UI
            (Default: off)
            """,
            # TODO: Add additional verification that checks that type has + or
            # .
            set(['on', 'off'])),


        Parameter(
            'spectrum',
            'use score to shade color items',
            # NOTE: BED 5 or higher
            'on'),

        Parameter(
            'scoreMax',
            'defines upper score limit that will receive graded scoring',
            # NOTE: BED 5 or higher, used with spectrum
            int),

        Parameter(
            'scoreMin',
            'defines lower score limit that will receive graded scoring',
            # NOTE: BED 5 or higher, used with spectrum
            int),

        Parameter(
            'thickDrawItem',
            """
            on|off; draw portions of items thicker to differentiate from
            other regions (Default: off)
            """,
            # NOTE: BED 8 or higher
            set(['on', 'off'])),

        # Less frequent used options

        Parameter(
            'bedNameLabel',
            """
            specifies an alternate label for the item name
            """,
            str),

        Parameter(
            'exonArrowDense',
            """
            on|off; display exon arrows even when the
            track is in dense mode (Default: off)
            """,
            # NOTE: BED 10 or higher
            set(['on', 'off'])),

        Parameter(
            'itemImagePath',
            """
            path suffix; specifies a URL or relative path to track
            that displays image on item details page
            """,
            str),

        Parameter(
            'itemBigImagePath',
            """
            path suffix; specifies a URL or relative path to track
            that provides link to bigger image if itemImagePath
            is specified
            """,
            str),

        Parameter(
            'nextExonText',
            """
            specifies an alternate label, when mousing over
            next exon arrows
            """,
            # NOTE: BED 10 or higher
            str),

        Parameter(
            'prevExonText',
            """
            specifies an alternate label, when mousing over
            previous exon arrows
            """,
            # NOTE: BED 10 or higher
            str),

        Parameter(
            'showTopScorers',
            """
            integer specifying a list of top-scoring items
            in genomic region on the item details page
            """,
            # NOTE: BED 5 or higher
            int),

    ]),

    'bigWig': OrderedDict((i.param, i) for i in [

        Parameter(
            'alwaysZero',
            """
            on|off; when autoScale is on, this setting
            ensures that y=0 is in the view (Default: off)
            """,
            set(['on', 'off'])),

        Parameter(
            'autoScale',
            """
            on|off; auto y-axis scaling to ensure that
            highest score in the current widow will
            peak at the top of the window (Default: off)
            """,
            set(['on', 'off'])),

        Parameter(
            'graphTypeDefault',
            """
            points|bars; signal can be graphed as points
            or bars (Default: bars)
            """,
            set(['points', 'bars'])),

        Parameter(
            'maxHeightPixels',
            """
            max:default:min; amount of vertical viewing space
            (Default: 100:16:8)
            """,
            validate.ColSV3),

        Parameter(
            'maxWindowToQuery',
            """
            a (large) positive number; if winEnd-winStart is larger, don't
            query items
            """,
            int),

        Parameter(
            'negateValues',
            """
            Negate the values in the wiggle, meaning that positive values
            become negative and vice-versa. This is useful for wiggles
            representing transcription or other activities on the Crick strand.
            Be aware that wiggles with negative values are drawn in altColor
            not color as positive values are.
            """,
            set(['on'])),

        Parameter(
            'smoothingWindow',
            """
            off|1-16; smoothing of graph using surrounding data,
            the numeric number determines how much surrounding data
            to use (Default: off_
            """,
            validate.off_or_int),

        Parameter(
            'transformFunc',
            'NONE|LOG; change representation scale (Default: NONE)',
            set(['NONE', 'LOG'])),

        Parameter(
            'viewLimits',
            "lower:upper; set default viewing range",
            # NOTE: autoScale must be off for this to work
            validate.ColSV2),

        Parameter(
            'viewLimitsMax',
            'lower:upper; unenforced -- suggested bounds of viewLimits',
            validate.ColSV2),

        Parameter(
            'windowingFunction',
            """
            mean|mean+whiskers|maximum|minimum; how to summarize signal
            data
            """,
            set(['maximum', 'mean', 'mean+whiskers', 'minimum'])),

        Parameter(
            'yLineMark',
            'float; position on y-axis to draw line across (Default: 0.0)',
            float),

        Parameter(
            'yLineOnOff',
            """on|off; draw y line at some fixed position set by yLineMark
            (Default: off)""",
            set(['on', 'off'])),

        Parameter(
            'gridDefault',
            "on|off; draw y=0.0 line (Default: off)",
            set(['on', 'off'])),

    ]),

    'bam':  OrderedDict((i.param, i) for i in [
        Parameter(
            'bamColorMode',
            """
            strand|gray|tag|off; coloring method (Default: strand)
            """,
            set(['strand', 'gray', 'tag', 'off'])),

        Parameter(
            'bamGrayMode',
            """
            aliQual|baseQual|unpaired; grayscale metric (Default: aliQual)
            """,
            # NOTE: bamColorMode gray
            set(['aliQual', 'baseQual', 'unpaired'])),

        Parameter(
            'aliQualRange',
            'min:max; shade alignment quals within this range (Default: 0:99)',
            validate.ColSV2),

        Parameter(
            'baseQualRange',
            'min:max; shade base quals within this range (Default: 0:40)',
            validate.ColSV2),

        Parameter(
            'bamColorTag',
            'optional tag for rgb color (Default: YC)',
            # NOTE: bamColorMode tag
            str),

        Parameter(
            'noColorTag',
            """
            placeholder, e.g. "."; if present don't offer option of setting
            color tag in track configuration page""",
            str),

        Parameter(
            'bamSkipPrintQualScore',
            """
            skip printing qual score in bam item details page; use "." to skip
            printing""",
            '.'),

        Parameter(
            'indelDoubleInsert',
            """
            on|off; highlight alignment gaps in target and query sequence
            with = (Default: 0ff)
            """,
            set(['on', 'off'])),

        Parameter(
            'indelQueryInsert',
            """
            on|off; highlight inserts in query sequence with
            orange or purple vertical lines (Default: off)
            """,
            set(['on', 'off'])),

        Parameter(
            'indelPolyA',
            """
            on|off; highlight poly-a tail with a vertical green line
            (Default: off)
            """,
            set(['on', 'off'])),

        Parameter(
            'minAliQual',
            'display only items above threshold quality score (Default: 0)',
            int),

        Parameter(
            'pairEndsByName',
            """
            if data has paired-end tags joined by name; use "."
            """,
            '.'),

        Parameter(
            'pairSearchRange',
            """
            search range to join pairs of tags (Default: 20000)
            """,
            int),

        Parameter(
            'showNames',
            "on|off; if off then don't display query names (Default: on)",
            set(['on', 'off'])),


        # Undocumented stuff, found from
        # https://lists.soe.ucsc.edu/pipermail/genome/2012-June/029615.html
        #
        # Uploading BAM files directly showed indels, but not when the same
        # file was included in a track hub.
        #
        # Currently not supported for bigBed
        Parameter(
            'baseColorUseSequence',
            'for BAM, try "lfExtra"; <extFile {seqTable} <extFile> / hgPcrResult / lfExtra / nameIsSequence / seq1Seq2 / ss >',
            set(['extFile', 'hgPcrResult', 'lfExtra', 'nameIsSequence', 'seq1Seq2', 'ss'])),

        Parameter(
            'baseColorDefault',
            '<diffBases/diffCodons/itemBases/itemCodons/genomicCodons>',
            set(['diffBases', 'diffCodons', 'itemBases', 'itemCodons', 'genomicCodons'])),

        Parameter(
            'showDiffBasesAllScales',
            '.',
            set(['.'])),





    ]),

    'vcfTabix': OrderedDict((i.param, i) for i in [
        # No Special Configuration for vcfTabix, uses maxWindowToDraw

    ]
    ),

}


composite_track_fields = OrderedDict((i.param, i) for i in [
    Parameter(
        'noInherit',
        """
        use this when subtracks of a different type are included in this
        composite """,
        'on'),

    Parameter(
        "allButtonPair",
        """
        Add all [=][-] buttons for selecting subtracks""",
        "on"),

    Parameter(
        "filterComposite",
        """
        ABC dimensions are represented by filter-like drop-downs.  Can be "on",
        or something like dimB=onlyOne in order to allow only a single
        selection for that dimension, overriding the default of dimB=multi""",
        str),

    Parameter(
        "dragAndDrop",
        """
        enable drag-and-drop subtracks""",
        "subtracks"),

    Parameter(
        'dimensions',
        "dimensionX=factor1 dimensionY=factor2 dimA=rep dimB=prot",
        # TODO: better verification
        str),

    Parameter(
        'sortOrder',
        'sort order for composite, e.g., cellType=+ factor=-',
        str),

    Parameter(
        'centerLabelsDense',
        'show subtrack labels even in dense view',
        set(['on', 'off'])),

])

view_track_fields = OrderedDict((i.param, i) for i in [
    Parameter(
        'view',
        "view tag, defined as a subgroup in parent composite",
        str),
    Parameter(
        'viewUi',
        'if on (default), displays view config settings',
        set(['on', 'off'])),
])

aggregate_track_fields = OrderedDict((i.param, i) for i in [
    Parameter(
        'aggregate',
        """
        transparentOverlay|solidOverlay|none; declares aggregate method
        """,
        set(['transparentOverlay', 'solidOverlay', 'none'])),

    Parameter(
        'showSubtrackColorOnUi',
        """
        on|off; show color associated with each track on the
        configuration page (Default: off)
        """,
        set(['on', 'off'])),

])

track_fields = OrderedDict((i.param, i) for i in [

    # Common Settings
    Parameter(
        'visibility',
        """hide|dense|squish|pack|full;default visibility for this track
        (Default: hide)""",
        set(['hide', 'dense', 'squish', 'pack', 'full'])
    ),

    Parameter(
        'html',
        """
        optional file containing description of a track in HTML and is path is
        relative to the path of trackDb file track is in.
        """,
        str),

    # Common Optional Settings
    Parameter(
        'boxedCfg',
        """
        on|off; places configuration controls within a box, much like
        multi-view controls have (Default: off)""",
        set(['on', 'off'])),

    Parameter(
        'color',
        'red, green, blue values 0 to 255; specifies primary color for items',
        validate.RGB),

    Parameter(
        'altColor',
        """
        red, green, blue values 0 to 255; specifies secondary color for
        items""",
        validate.RGB),

    Parameter(
        'chromosomes',
        """
        chr1, chr3; only these chroms have data for this track, the system
        displays message that there is no data on other chroms
        """,
        validate.CSV),

    Parameter(
        'dataVersion',
        """
        display a version statement for this track on configuration and details
        pages, supports limited HTML
        """,
        str),

    Parameter(
        'directUrl',
        """
        url;
        link image to alternative details page using these fields in order:
        item name, chromosome name, chromosome start position, chromosome end
        position, track name, database name;
        """,
        # TODO: better verification and information
        str),

    Parameter(
        'otherDb',
        'declare the other species/assembly used in the pairwise alignments',
        # TODO: determine if functionality works for bigBed
        str),

    Parameter(
        'pennantIcon',
        """
        icon [html [tip]]; Displays icon next to track in "parade" of tracks
        found in hgTracks config. Html is an optional page describing the icon
        and path can be relative to the track or absolute. Tip is an optional
        "quoted string" that is seen when the user hovers over the icon.
        """,
        str),

    Parameter(
        'priority',
        """
        used to order this track within its track group and within the browser
        image, tracks of the same priority are sorted alphabetically by short
        label.  (Default: 0)
        """,
        float),

    Parameter(
        'url',
        'url; place an external link on the details page',
        # TODO: improved verification and information on coding
        str),

    Parameter(
        'url2',
        'url; place an additional external link on the details page',
        # TODO: improved verification and information on coding
        str),

    Parameter(
        'urlLabell',
        """
        used with url setting and provides text for link (Default: "outside
        link:")
        """,
        str),

    Parameter(
        'url2Label',
        """
        used with url2 setting and provides text for link (Default: "outside
        link")
        """,
        str),


    # Across datatype Settings

    Parameter(
        'maxWindowToDraw',
        """
        a (large) positive number; if winEnd-winStart is larger than threshold
        forces track to dense mode""",
        int),


    Parameter(
        'configurable',
        """
        on|off; on for all track types that allow configuration.  If set to
        off, configuration is blocked.  This setting is most useful in
        composites and views where some subtracks should be configurable and
        others not. (Default: on)""",
        set(['on', 'off'])),


])



import inspect
import sys
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict


class HubComponent(object):
    """
    Base class for various track hub components.  Several methods must be
    overridden by subclasses.
    """
    def __init__(self):
        self.children = []
        self.parent = None

    def _render(self):
        """
        Renders the object to file.  Must be overridden by subclass.

        Can return None if nothing to be done for the subclass.
        """
        raise NotImplementedError(
            "%s: subclasses must define their own _render() method"
            % self.__class__.name__)

    def validate(self):
        """
        Runs validation, raising exceptions as needed.  Must be overridden by
        subclass.
        """
        raise NotImplementedError(
            "%s: subclasses must define their own validate() method"
            % self.__class__.__name__)

    def add_child(self, child):
        """
        Adds self as parent to child, and then adds child.
        """
        child.parent = self
        self.children.append(child)
        return child

    def add_parent(self, parent):
        """
        Adds self as child of parent, then adds parent.
        """
        parent.add_child(self)
        self.parent = parent
        return parent

    def root(self, cls=None, level=0):
        """
        Returns the top-most HubComponent in the hierarchy.

        If `cls` is not None, then return the top-most attribute HubComponent
        that is an instance of class `cls`.

        For a fully-constructed track hub (and `cls=None`), this should return
        a a Hub object for every component in the hierarchy.
        """
        if cls is None:
            if self.parent is None:
                return self, level
        else:
            if isinstance(self, cls):
                if not isinstance(self.parent, cls):
                    return self, level

        if self.parent is None:
            return None, None

        return self.parent.root(cls, level - 1)

    def leaves(self, cls, level=0, intermediate=False):
        """
        Returns an iterator of the HubComponent leaves that are of class `cls`.

        If `intermediate` is True, then return any intermediate classes as
        well.
        """
        if intermediate:
            if isinstance(self, cls):
                yield self, level
        elif len(self.children) == 0:
            if isinstance(self, cls):
                yield self, level
            else:
                raise StopIteration

        for child in self.children:
            for leaf, _level in child.leaves(cls, level + 1, intermediate=intermediate):
                    yield leaf, _level

    def render(self):
        """
        Renders the object to file, returning a list of created files.

        Calls validation code, and, as long as each child is also a subclass of
        :class:`HubComponent`, the rendering is recursive.
        """
        self.validate()
        created_files = OrderedDict()
        this = self._render()
        if this:
            created_files[repr(self)] = this
        for child in self.children:
            created_files[repr(child)] = child.render()
        return created_files



import os
import sys
import json
import base
from hub import Hub

_here = __file__

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

def auto_track_url(track):
    """
    Automatically sets the bigDataUrl for `track`.

    Requirements:

        * the track must be fully connected, such that its root is a Hub object
        * the root Hub object must have the Hub.url attribute set
        * the track must have the `local_fn` attribute set
    """
    hub = track.root(cls=Hub)

    if hub is None:
        raise ValueError(
            "track is not fully connected because the root is %s" % repr(hub))
    if hub.url is None:
        raise ValueError( "hub.url is not set")
    if track.local_fn is None:
        raise ValueError("track.local_fn is not set")




def show_rendered_files(results_dict):
    """
    Parses a nested dictionary returned from :meth:`Hub.render` and just prints
    the resulting files.
    """
    for k, v in results_dict.items():
        if isinstance(v, basestring):
            print "rendered file: %s (created by: %s)" % (v, k)
        else:
            show_rendered_files(v)
    return


def print_rendered_results(results_dict):
    """
    Pretty-prints the rendered results dictionary.

    Rendered results can be multiply-nested dictionaries; this uses JSON
    serialization to print a nice representation.
    """
    class _HubComponentEncoder(json.JSONEncoder):
        def default(self, o):
            if isinstance(o, base.HubComponent):
                return repr(o)
            return json.JSONEncoder.default(self, o)
    formatted = json.dumps(results_dict, indent=4, cls=_HubComponentEncoder)
    # the returned string contains lines with trailing spaces, which causes
    # doctests to fail.  So fix that here.
    for s in formatted.splitlines():
        print s.rstrip()


def data_dir():
    """
    Returns the data directory that contains example files for tests and
    documentation.
    """
    return os.path.join(os.path.dirname(_here), 'test', 'data')


def example_bigbeds():
    """
    Returns list of example bigBed files
    """
    hits = []
    d = data_dir()
    for fn in os.listdir(d):
        fn = os.path.join(d, fn)
        if os.path.splitext(fn)[-1] == '.bigBed':
            hits.append(os.path.abspath(fn))
    return hits


def example_bigwigs():
    """
    Returns list of bigWig example files
    """
    hits = []
    d = data_dir()
    for fn in os.listdir(d):
        fn = os.path.join(d, fn)
        if os.path.splitext(fn)[-1] == '.bw':
            hits.append(os.path.abspath(fn))
    return hits



import os
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
from validate import ValidationError
from hub import Hub
from base import HubComponent


class GenomesFile(HubComponent):
    def __init__(self, genome=None):
        """
        Represents the genomes file on disk.  Can contain multiple
        :class:`Genome` objects, each of which represent a stanza in this
        file.

        The file ultimately created (with the self.render() method) will be
        determined by the parent Hub's `genome_filename` attribute.  By
        default, this is the hub name, plus ".genomes.txt"
        """
        HubComponent.__init__(self)
        self._local_fn = None
        self._remote_fn = None
        self.genomes = []
        if genome is None:
            genome = []
        for genome in genome:
            self.add_genome(genome)

    @property
    def hub(self):
        hub, level = self.root(Hub)
        if level is None:
            return None
        if level != -1:
            raise ValueError(
                "Found a hub at %s levels away -- needs to be -1" % level)
        return hub

    @property
    def local_fn(self):
        if self._local_fn is not None:
            return self._local_fn
        if self.hub is None:
            return None
        return os.path.join(
            os.path.dirname(self.hub.local_fn),
            self.hub.hub + '.genomes.txt')

    @local_fn.setter
    def local_fn(self, fn):
        self._local_fn = fn

    @property
    def remote_fn(self):
        if self._remote_fn is not None:
            return self._remote_fn
        if self.hub is None:
            return None
        return os.path.join(
            os.path.dirname(self.hub.remote_fn),
            self.hub.hub + '.genomes.txt')

    @remote_fn.setter
    def remote_fn(self, fn):
        self._remote_fn = fn

    def add_genome(self, genome):
        self.add_child(genome)
        self.genomes = self.children

    def __str__(self):
        s = []
        for genome in self.genomes:
            s.append(str(genome))
        return '\n'.join(s) + '\n'

    def _render(self):
        """
        Renders the children Genome objects to file
        """
        fout = open(self.local_fn, 'w')
        fout.write(str(self))
        fout.close()
        return fout.name

    def validate(self):
        if len(self.children) == 0:
            raise ValueError(
                "No defined Genome objects to use")



from hub import Hub
import helpers
from genomes_file import GenomesFile
from genome import Genome
from trackdb import TrackDb
from track import Track, SubGroupDefinition, CompositeTrack, \
    ViewTrack, SuperTrack, AggregateTrack
from version import version as __version__


def default_hub(hub_name, genome, short_label, long_label, email):
    """
    Returns a fully-connected set of hub components using default filenames.
    """
    hub = Hub(
        hub=hub_name,
        short_label=short_label,
        long_label=long_label,
        email=email)
    genome = Genome(genome)
    genomes_file = GenomesFile()
    trackdb = TrackDb()
    hub.add_genomes_file(genomes_file)
    genomes_file.add_genome(genome)
    genome.add_trackdb(trackdb)
    return hub, genomes_file, genome, trackdb



import tempfile
import os
from fabric.api import local, settings, run, abort, cd, env, hide, puts
from fabric.contrib.console import confirm
from fabric.colors import green, yellow
import track


def upload_file(host, user, local_fn, remote_fn, port=22,
                rsync_options='-azvrL --progress', run_local=False,
                symlink=False, symlink_dir='staging'):
    results = []
    puts('\n\n' + green(local_fn) + ' -> ' +  yellow(remote_fn))
    if symlink:
        symlink_dest = os.path.join(symlink_dir, remote_fn)
        os.system('mkdir -p %s' % (os.path.dirname(symlink_dest)))
        remote_dir = os.path.dirname(remote_fn)
        os.system('ln -sf %s %s' % (os.path.abspath(local_fn), symlink_dest))
        local_fn = symlink_dest

    if run_local:
        if not os.path.exists(os.path.dirname(remote_fn)):
            remote_dir = os.path.dirname(remote_fn)
            if remote_dir:
                results.append(local('mkdir -p %s' % remote_dir))

        results.append(local('rsync %(rsync_options)s %(local_fn)s %(remote_fn)s' % locals()))
        return results
    env.host_string = host
    env.user = user
    env.port = port

    remote_dir = os.path.dirname(remote_fn)

    with settings(warn_only=True):
        result = run('ls %s' % remote_dir)

    if result.failed:
        run('mkdir -p %s' % remote_dir)

    rsync_template = \
        'rsync %(rsync_options)s %(local_fn)s %(user)s@%(host)s:%(remote_fn)s'
    with settings(warn_only=True):
        results.append(local(rsync_template % locals()))
    return results


def upload_hub(host, user, hub, port=22, rsync_options='-azvrL --progress',
               run_local=False, symlink=False, symlink_dir='staging'):
    kwargs = dict(host=host, user=user, port=port, rsync_options=rsync_options,
                  run_local=run_local, symlink=symlink, symlink_dir=symlink_dir)
    print kwargs
    results = []

    # First the hub file:
    results.extend(
        upload_file(
            local_fn=hub.local_fn,
            remote_fn=hub.remote_fn,
            **kwargs)
    )

    # Then the genomes file:
    print hub.genomes_file.local_fn
    results.extend(
        upload_file(
            local_fn=hub.genomes_file.local_fn,
            remote_fn=hub.genomes_file.remote_fn,
            **kwargs)
    )

    # then the trackDB file:
    for g in hub.genomes_file.genomes:
        results.extend(
            upload_file(
                local_fn=g.trackdb.local_fn,
                remote_fn=g.trackdb.remote_fn,
                **kwargs
            )
        )
    # and finally any HTML files:
    for t, level in hub.leaves(track.CompositeTrack, intermediate=True):
        print repr(t)
        if t._html:
            results.extend(
                upload_file(
                    local_fn=t._html.local_fn,
                    remote_fn=t._html.remote_fn,
                    **kwargs)
            )
    return results


def upload_track(host, user, track, port=22, rsync_options='-azvrL --progress',
                 run_local=False):
    kwargs = dict(host=host, user=user, local_fn=track.local_fn,
                  remote_fn=track.remote_fn, rsync_options=rsync_options,
                  run_local=run_local)
    results = upload_file(**kwargs)
    if track.tracktype == 'bam':
        kwargs['local_fn'] += '.bai'
        kwargs['remote_fn'] += '.bai'
        results.extend(upload_file(**kwargs))

    if track.tracktype == 'vcfTabix':
        kwargs['local_fn'] += '.tbi'
        kwargs['remote_fn'] += '.tbi'
        results.extend(upload_file(**kwargs))
    return results



import os
import subprocess
import pybedtools
from trackhub import helpers

data_dir = helpers.data_dir()

chromsizes = pybedtools.genome_registry.dm3.euchromatic
g = pybedtools.chromsizes_to_file(chromsizes)

# Make some randomized bigBed files
for i in range(3):
    x = pybedtools.BedTool(
        "chr2L 0 10000000", from_string=True)\
        .window_maker(w=1000 * (i + 1))\
        .shuffle(g=g, seed=i)\
        .sort()

    out = os.path.join(data_dir, 'random-dm3-%s.bigBed' % i)

    cmds = [
        'bedToBigBed',
        x.fn,
        g,
        out]
    p = subprocess.Popen(cmds)
    p.communicate()


# make some sine waves for bigWigs
import numpy as np
def sine(factor):

    for chrom, size in chromsizes.items():
        _, size = size
        x = np.arange(0, size, 10000)
        y = np.sin(x / factor / np.pi) + np.random.random(len(x)) / 3
        for xi, yi in zip(x, y):
            yield pybedtools.create_interval_from_list(map(str, [
                chrom, xi, xi + 1000, yi]))

for f in [1000., 10000.]:
    x = pybedtools.BedTool(sine(f)).saveas(os.path.join(data_dir, 'sine-dm3-%d.bedgraph' % f))
    out = x.fn + '.bw'

    cmds = [
        'bedGraphToBigWig',
        x.fn,
        g,
        out]
    p = subprocess.Popen(cmds)
    p.communicate()
    os.unlink(x.fn)




from trackhub import upload
import os
from trackhub import Hub, GenomesFile, Genome, Track, CompositeTrack, \
    TrackDb, ViewTrack, SuperTrack, AggregateTrack


class TestUpload(object):
    def setup(self):
        self.hub = Hub(
            hub='example_hub',
            short_label='example hub',
            long_label='an example hub for testing',
            email='none@example.com')
        self.genomes_file = GenomesFile()
        self.genome = Genome('dm3')
        self.trackdb = TrackDb()

        self.tracks = [
            Track(
                name='track1',
                tracktype='bam',
                local_fn='data/track1.bam'
            ),
            Track(
                name='track2',
                tracktype='bigWig',
                local_fn='data/track2.bigwig',
            ),
        ]
        self.hub.add_genomes_file(self.genomes_file)
        self.genomes_file.add_genome(self.genome)
        self.genome.add_trackdb(self.trackdb)
        self.trackdb.add_tracks(self.tracks)


    def test_upload(self):
        self.hub.remote_fn = os.path.join(
            'uploaded_version',
            self.hub.remote_fn)
        self.hub.render()
        upload.upload_hub(
            'localhost',
            None,
            self.hub,
            symlink=True,
            symlink_dir='staging',
            run_local=True,)
        for t, level in self.hub.leaves(Track):
            upload.upload_track(
                track=t, host='localhost', user=None, run_local=True)



from textwrap import dedent


class ValidationError(Exception):
    pass


def validator(*example):
    def wrapper(func):

        # Try running validation on the function's own example . . . it better
        # work!
        example_string = []
        try:
                for ex in example:
                    func(ex)
                    example_string.append(ex)
                example_string = 'Example value(s): %s' \
                    % (" or ".join('%r' % i for i in example_string))

        except Exception as e:
            raise ValueError(
                'Error validating example (func=%s, example=%r)! '
                '\nOriginal error:\n\t%s: %s'
                % (func.func_name, example, e.__class__.__name__, e.message))

        class Validator(object):
            """
            Class to wrap a function and display an example value.
            """
            def __call__(self, v):
                try:
                    return func(v)
                except Exception as e:
                    raise ValidationError(
                        '%s; %s'
                        % (e.message, example_string))

            def __str__(self):
                return "<Validator [%s] at %s> sample: %s" \
                    % (func.func_name, id(func), example)

        return Validator()
    return wrapper


class Parameter(object):
    """
    Represents a generic parameter. Validates based on the provided possible
    values, e.g.::

        >>> Parameter("name", "the name to use", str).validate("asdf")
        True

        >>> Parameter("name", "the name to use", int).validate(999)
        False
    """
    def __init__(self, param, desc, values):
        self.param = param
        self.desc = " ".join(dedent(desc).splitlines(False))
        self.values = values

    def __str__(self):
        return '<%s "%s" at %s>' \
            % (self.__class__.__name__, self.param, id(self))

    def validate(self, value):

        if isinstance(self.values, set):
            if value in self.values:
                return True

        if isinstance(self.values, type):
            if isinstance(value, self.values):
                return True
            return False

        if hasattr(self.values, '__call__'):
            return self.values(value)

        elif value == self.values:
            return True


@validator("tag=value")
def key_val(v):
    try:
        assert '=' in v
        items = v.split()
        assert len(items) == v.count('=')
        return True

    except AssertionError:
        return False


@validator("a,b,c")
def CSV(v):
    #TODO: is a one-item list "chr1," or "chr1"?
    if isinstance(v, basestring):
        return True


@validator("a:b:c")
def ColSV3(v):
    nvalues = 3
    if not isinstance(v, basestring):
        raise ValueError('not a string')
    vs = v.split(':')
    assert len(vs) == nvalues
    return True


@validator("a:b")
def ColSV2(v):
    nvalues = 2
    if not isinstance(v, basestring):
        raise ValueError('not a string')
    vs = v.split(':')
    assert len(vs) == nvalues
    return True


@validator("128,0,255")
def RGB(v):
    if ' ' in v:
        raise ValueError('Space in RGB tuple')
    if "." in v:
        raise ValueError('"." in RGB tuple')
    assert isinstance(v, basestring), "RGB tuple is not a string"
    if ',' not in v:
        raise ValueError("no commas in RGB tuple")
    rgb = v.split(',')
    assert len(rgb) == 3, "RGB tuple does not have 3 values"
    try:
        rgb = map(int, rgb)
    except ValueError:
        raise ValueError('RGB tuple does not contain ints')
    for i in rgb:
        assert (0 <= i <= 255)
    return True


@validator("128,0,0 90,90,5")
def RGBList(v):
    rgbs = v.split(' ')
    assert len(rgbs) == 2, "RGBList not a space-separated list of RGB tuples"
    for i in rgbs:
        RGB(i)
    return True


@validator("off", 1)
def off_or_int(v):
    try:
        int(v)
        return True
    except ValueError:
        if v == 'off':
            return True
        raise



import os
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
from validate import ValidationError
from base import HubComponent


class Genome(HubComponent):
    def __init__(self, genome, trackdb=None, genome_file_obj=None):
        """
        Represents a 2-line genome stanza within a "genomes.txt" file.

        The file itself is represented by a :class:`GenomesFile` object.
        """
        HubComponent.__init__(self)
        self.genome = genome
        self.trackdb = None
        if trackdb is not None:
            self.add_trackdb(trackdb)
        if genome_file_obj:
            self.add_parent(genome_file_obj)

    @property
    def genome_file_obj(self):
        try:
            return self.parent
        except AttributeError:
            return None

    def add_trackdb(self, trackdb):
        self.children = []
        self.add_child(trackdb)
        self.trackdb = self.children[0]

    def __str__(self):
        try:
            self.validate()
        except ValidationError:
            return "Unconfigured <Genome> object"
        s = []
        s.append('genome %s' % self.genome)
        trackdb_relpath = os.path.relpath(self.trackdb.local_fn, 
                                          start = os.path.dirname(
                                                    self.parent.parent.local_fn)) #local_fn of hub file
        s.append('trackDb %s' % trackdb_relpath)
        return '\n'.join(s) + '\n'

    def validate(self):
        if len(self.children) == 0:
            raise ValidationError(
                "No TrackDb objects provided")
        if self.trackdb is None:
            raise ValidationError("No TrackDb objects provided")

    def _render(self):
        """
        No file is created from a Genome object -- only from its parent
        GenomesFile object.
        """
        pass



from nose.tools import assert_raises
from trackhub import Hub, GenomesFile, Genome, Track, CompositeTrack, \
    TrackDb, ViewTrack, SuperTrack, AggregateTrack
import os


class TestComponents(object):
    def setup(self):
        self.hub = Hub(
            hub='example_hub',
            short_label='example hub',
            long_label='an example hub for testing',
            email='none@example.com')
        self.genomes_file = GenomesFile()
        self.genome = Genome('dm3')
        self.trackdb = TrackDb()

        self.tracks = [
            Track(name='track1', tracktype='bam'),
            Track(name='track2', tracktype='bigWig'),
        ]

    def CONNECT(self):
        """
        Connect the components together. The default setup creates the objects
        but does not connect them.
        """
        self.hub.add_genomes_file(self.genomes_file)
        self.genomes_file.add_genome(self.genome)
        self.genome.add_trackdb(self.trackdb)
        self.trackdb.add_tracks(self.tracks)

    def DISCONNECT(self):
        """
        Re-run the setup, which results in unconnected components. Run
        CONNECT() to connect them up.
        """
        self.setup()

    def test_self_connection(self):
        """
        meta test: make sure the test class's connect/disconnect is working
        """
        assert self.hub.genomes_file is None

        self.CONNECT()
        assert self.hub.genomes_file is self.genomes_file

        self.DISCONNECT()
        assert self.hub.genomes_file is None

    # Filenames ---------------------------------------------------------------

    def test_hub_fns(self):
        # Default unconnected
        assert self.hub.local_fn == 'example_hub.hub.txt'
        assert self.hub.remote_fn == 'example_hub.hub.txt'
        assert self.hub.url is None

        # Connecting components should not change hub
        self.CONNECT()
        assert self.hub.local_fn == 'example_hub.hub.txt'
        assert self.hub.remote_fn == 'example_hub.hub.txt'
        assert self.hub.url is None
        self.DISCONNECT()

        # set local/remote dir and fns should follow
        self.hub.local_dir = '/data/hubs'
        self.hub.remote_dir = '/remote/hubs'
        assert self.hub.local_fn == '/data/hubs/example_hub.hub.txt', \
            self.hub.local_fn
        assert self.hub.remote_fn == '/remote/hubs/example_hub.hub.txt', \
            self.hub.remote_fn

        # resetting them should not change
        self.hub.local_dir = None
        self.hub.remote_dir = None
        assert self.hub.local_fn == 'example_hub.hub.txt'
        assert self.hub.remote_fn == 'example_hub.hub.txt'

        # changing local_fn or remote_fn should stick -- and should override
        # any set local/remote dir
        self.hub.local_fn = 'dummy.txt'
        self.hub.remote_fn = 'remote.txt'
        self.hub.local_dir = 'no_dir'
        self.hub.remote_dir = 'remote_dir'
        assert self.hub.local_fn == 'dummy.txt'
        assert self.hub.remote_fn == 'remote.txt'

    def test_genome_file_fns(self):
        assert_raises(AttributeError, getattr, self.genomes_file, 'url')
        # When unconnected, filenames should be None
        assert self.genomes_file.local_fn is None
        assert self.genomes_file.remote_fn is None

        #...though you can set them manually
        self.genomes_file.local_fn = 'local.genomes'
        assert self.genomes_file.local_fn == 'local.genomes'
        self.genomes_file.local_fn = None

        self.CONNECT()
        assert self.genomes_file.local_fn == 'example_hub.genomes.txt'

        # when connected, overriding works
        self.genomes_file.local_fn = 'local.genomes'
        assert self.genomes_file.local_fn == 'local.genomes'
        self.genomes_file.local_fn = None

        # disconnecting brings it back to None
        self.DISCONNECT()
        assert self.genomes_file.local_fn is None

        # set the hub's local_dir; genomes_file should follow.
        self.CONNECT()
        self.hub.local_dir = 'local/'
        assert self.genomes_file.local_fn == 'local/example_hub.genomes.txt'

        # what happens if the hub's local FN is changed?
        self.hub.local_fn = 'localhub/hub.txt'
        assert self.genomes_file.local_fn == 'localhub/example_hub.genomes.txt'

    def test_genome_fns(self):
        # should be easy -- filenames should raise attribute errors
        assert_raises(AttributeError, getattr, self.genome, 'local_fn')
        assert_raises(AttributeError, getattr, self.genome, 'remote_fn')
        assert_raises(AttributeError, getattr, self.genome, 'url')

    def test_trackdb_fns(self):
        assert_raises(AttributeError, getattr, self.trackdb, 'url')

        # when unconnected, no defaults
        assert self.trackdb.local_fn is None
        assert self.trackdb.remote_fn is None

        self.CONNECT()
        assert self.trackdb.local_fn == 'dm3/trackDb.txt'
        assert self.trackdb.remote_fn == 'dm3/trackDb.txt'

        # setting the local dir on the hub trickles down
        self.hub.local_dir = 'localdir'
        assert self.trackdb.local_fn == 'localdir/dm3/trackDb.txt'

        # setting local_fn overrides
        self.trackdb.local_fn = 'mytrackdb.txt'
        assert self.trackdb.local_fn == 'mytrackdb.txt', self.trackdb.local_fn

        # ...and back to None to reset
        self.trackdb.local_fn = None
        assert self.trackdb.local_fn == 'localdir/dm3/trackDb.txt'

        # genomes_file fn overrides
        self.genomes_file.local_fn = 'anotherdir/genomes.txt'
        assert self.trackdb.local_fn == 'anotherdir/dm3/trackDb.txt'

        # reset parent hub and genomes file to get back to the default
        self.genomes_file.local_fn = None
        self.hub.local_dir = None
        assert self.trackdb.local_fn == 'dm3/trackDb.txt'

    def test_track_fns(self):
        for track in self.tracks:
            assert track.local_fn is None
            assert track.remote_fn is None
            assert track.url is None

        self.CONNECT()
        # local fns should still be None
        for track in self.tracks:
            assert track.local_fn is None

        # remote_fn is relative to the hub's remote_fn
        assert self.tracks[0].remote_fn == 'dm3/track1.bam'
        assert self.tracks[1].remote_fn == 'dm3/track2.bigWig'

        self.hub.remote_fn = '/var/www/hubs/hub.txt'
        self.hub.url = 'http://example.com/hubs/hub.txt'

        # URL is relative to the trackDb
        assert self.tracks[0].url == 'track1.bam'


    def test_track_creation(self):
        track = Track(name='track0', tracktype='bam', local_fn='t0.bam')
        assert track.local_fn == 't0.bam'



import ez_setup
ez_setup.use_setuptools()

import os
import sys
from setuptools import setup

version_py = os.path.join(os.path.dirname(__file__), 'trackhub', 'version.py')
version = open(version_py).read().strip().split('=')[-1].replace('"','')

long_description = """
Create and manage UCSC track hubs from Python
"""

install_requires = ['fabric']
if (sys.version_info[0] == 2) and (sys.version_info[1] < 7):
    install_requires.append('ordereddict')

setup(
    name="trackhub",
    version=version,
    install_requires=[i.strip() for i in open('requirements.txt')],
    packages=['trackhub',
              'trackhub.test',
              'trackhub.test.data',
              'trackhub.scripts'],
    author="Ryan Dale",
    description=long_description,
    long_description=long_description,
    url="http://github.com/daler/trackhub",
    package_data = {'trackhub':["test/data/*"]},
    package_dir = {"trackhub": "trackhub"},
    license = 'MIT',
    #scripts = ['trackhub/scripts/example_script.py'],
    author_email="dalerr@niddk.nih.gov",
    classifiers=['Development Status :: 4 - Beta'],
)



import os
from validate import ValidationError
from base import HubComponent
from genomes_file import GenomesFile
from genome import Genome
import track as _track

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

class TrackDb(HubComponent):
    def __init__(self, tracks=None):
        """
        Represents the file containing one or more Track objects (which each
        represent a stanza).
        """
        HubComponent.__init__(self)
        if tracks is None:
            tracks = []

        self._tracks = []
        for track in tracks:
            self.add_track(track)

        self._local_fn = None
        self._remote_fn = None

    @property
    def hub(self):
        return self.root(cls=Hub)

    @property
    def genomes_file(self):
        genomes_file, level = self.root(GenomesFile)
        if level is None:
            return None
        if level != -2:
            raise ValueError("GenomesFile is level %s, not -2" % level)
        return genomes_file

    @property
    def genome(self):
        genome, level = self.root(Genome)
        if level is None:
            return None
        if level != -1:
            raise ValueError('Genome is level %s, not -1' % level)
        return genome

    @property
    def local_fn(self):
        if self._local_fn is not None:
            return self._local_fn

        if self.genome is None:
            return None

        if self.genomes_file is None:
            return None

        else:
            return os.path.join(os.path.dirname(self.genomes_file.local_fn),
                                self.genome.genome, 'trackDb.txt')

    @local_fn.setter
    def local_fn(self, fn):
        self._local_fn = fn

    @property
    def remote_fn(self):
        if self._remote_fn is not None:
            return self._remote_fn

        if self.genome is None:
            return None

        if self.genomes_file is None:
            return None

        else:
            return os.path.join(os.path.dirname(self.genomes_file.remote_fn),
                                self.genome.genome, 'trackDb.txt')

    @remote_fn.setter
    def remote_fn(self, fn):
        self._remote_fn = fn

    def add_tracks(self, track):
        """
        Add a track or iterable of tracks.

        :param track:
            Iterable of :class:`Track` objects, or a single :class:`Track`
            object.
        """
        if isinstance(track, _track.BaseTrack):
            self.add_child(track)
            self._tracks.append(track)
        else:
            for t in track:
                self.add_child(t)
                self._tracks.append(t)

    @property
    def tracks(self):
        return [i for i, level in self.leaves(_track.Track)]

    def add_genome(self, genome):
        self.add_parent(genome)

    def __str__(self):
        s = []
        for track in self._tracks:
            s.append(str(track) + '\n')
        return '\n'.join(s)

    def validate(self):
        if len(self.children) == 0:
            raise ValueError("No Track objects specified")

    def _render(self):
        dirname = os.path.dirname(self.local_fn)

        if not os.path.exists(dirname):
            os.makedirs(dirname)

        fout = open(self.local_fn, 'w')
        fout.write(str(self))
        fout.close()
        return fout.name



import sys
import os
import re
from validate import ValidationError
from base import HubComponent
import hub
import trackdb
import genomes_file
import genome
import constants

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict


TRACKTYPES = ['bigWig', 'bam', 'bigBed', 'vcfTabix', None]


def _check_name(name):
    regex = re.compile('[^a-zA-Z0-9-_]')
    if regex.search(name):
        raise ValueError('Non-alphanumeric character in name "%s"' % name)


class ParameterError(Exception):
    pass


class SubGroupDefinition(object):
    def __init__(self, name, label, mapping, default="none"):
        """
        Represents a subgroup line in a composite track.

        Instances of this class are provided to a composite track in order to
        define options for the subtracks' groups.

        :param name:
            String; name for the subgroup (e.g., "celltype").

        :param label:
            String; the label that will be displayed (e.g., "Cell_Type")

        :param mapping:
            Dictionary of {tag: title}, where `tag` will be how subtracks
            access this group and `title` is how it will be displayed in the
            browser, e.g.::

                {
                  "ES": "Embryonic stem cell",
                 "MEF": "Mouse embryonic fibroblast"
                }

            Upon appending this SubGroupDefinition to a composite track, the
            options for the subtracks' subgroups are required to come from the
            keys in the mapping.  Continuing the example, "celltype=ES" would
            be a valid subgroup for a subtrack, but "celltype=other" would not
            since it's not in the mapping dict above.

        :param default:
            String; value to be used by subtracks if they don't explicitly
            define this subgroup. Continuing the example, if a subtrack didn't
            specify the "celltype" subgroup, then by default a "celltype=none"
            value will be added.  This is necessary because subtracks must
            define a value for all groups.
        """
        self.name = name
        self.label = label
        self.mapping = mapping
        self.default = default

    def __str__(self):
        s = []
        s.append(self.name)
        s.append(self.label)
        s.extend('%s=%s' % (k, v) for k, v in self.mapping.items())
        return ' '.join(s)


class BaseTrack(HubComponent):

    # Dictionary where keys are parameter names (e.g., "color") and values are
    # Parameter objects.  These are defined in the constants module.
    params = OrderedDict()
    params.update(constants.track_field_order)
    params.update(constants.track_fields)

    # For a plain 'ol Track, there's nothing specific.  But CompositeTracks and
    # ViewTracks can have addtional specific parameters (e.g., filterComposite
    # for CompositeTracks) that need to be handled separately.  Make some space
    # for that here.
    specific_params = OrderedDict()

    def __init__(self, name, tracktype=None, short_label=None,
                 long_label=None, subgroups=None, local_fn=None,
                 remote_fn=None, html_string=None, **kwargs):
        """
        Represents a single track stanza.

        :param name: String; name of the track

        :param tracktype: String; type of the track (e.g., "bam")

        :param url: String; full URL for the track (i.e., bigDataUrl)

        :param short_label: String; used for the left-hand side track label

        :param long_label:
            String; used for the longer middle labels; if None will copy
            short_label

        :param subgroups:

            A dictionary of `{name: tag}` where each `name` is the name of
            a SubGroupDefinition in a parent :class:`CompositeTrack` and each
            `tag` is a key in the SubGroupDefinition.mapping dictionary.  They
            end up looking like this in the string representation::

                subGroups view=aln celltype=ES

        :param local_fn:
            String; Local path to the file (used for uploading)

        :param remote_fn:
            String; path to upload the file to, over rsync and ssh.
        """
        HubComponent.__init__(self)
        _check_name(name)
        self.name = name
        self.tracktype = tracktype
        if short_label is None:
            short_label = name
        self.short_label = short_label
        if long_label is None:
            long_label = short_label
        self.long_label = long_label

        self._local_fn = local_fn
        self._remote_fn = remote_fn
        self.html_string = html_string
        self.subgroups = {}
        self.add_subgroups(subgroups)

        # Convert pythonic strings to UCSC versions
        kwargs['track'] = name
        kwargs['type'] = tracktype
        kwargs['longLabel'] = kwargs.get('longLabel', long_label)
        kwargs['shortLabel'] = kwargs.get('shortLabel', short_label)

        self.kwargs = kwargs

        self._orig_kwargs = kwargs.copy()

    @property
    def _html(self):
        if not self.html_string:
            return None
        _html = HTMLDoc(self.html_string)
        _html.add_parent(self)
        return _html

    @property
    def trackdb(self):
        return self.root(trackdb.TrackDb)[0]

    @property
    def hub(self):
        return self.root(hub.Hub)[0]

    @property
    def local_fn(self):
        if self._local_fn is not None:
            return self._local_fn
        return None

    @local_fn.setter
    def local_fn(self, fn):
        self._local_fn = fn

    @property
    def remote_fn(self):
        if self._remote_fn is not None:
            return self._remote_fn
        # If remote_fn hasn't been assigned then make one automatically based
        # on the track name and the trackhub's remote_fn (which, by the way,
        # acts similarly, deferring up to the genomes_file.remote_fn . . . and
        # so on up to the hub's remote_fn)
        if self.trackdb:
                return os.path.join(
                    os.path.dirname(self.trackdb.remote_fn),
                    self.name + '.' + self.tracktype.split(' ')[0])
        return None

    @remote_fn.setter
    def remote_fn(self, fn):
        self._remote_fn = fn

    @property
    def tracktype(self):
        return self._tracktype

    @tracktype.setter
    def tracktype(self, tracktype):
        """
        When setting the track type, the valid parameters for this track type
        need to be set as well.
        """
        self._tracktype = tracktype
        if tracktype is not None:
            if 'bed' in tracktype.lower():
                tracktype = 'bigBed'
            elif 'wig' in tracktype.lower():
                tracktype = 'bigWig'
        self.params.update(constants.track_typespecific_fields[tracktype])

    def add_trackdb(self, trackdb):
        """
        Attach this track to a parent TrackDb object.
        """
        self.add_parent(trackdb)

    def add_params(self, **kw):
        """
        Add [possibly many] parameters to the track.

        Parameters will be checked against known UCSC parameters and their
        supported formats.

        E.g.,

        add_params(color='128,0,0', visibility='dense')

        """
        for k, v in kw.items():
            if (k not in self.params) and (k not in self.specific_params):
                raise ParameterError('"%s" is not a valid parameter for %s'
                                     % (k, self.__class__.__name__))
            try:
                self.params[k].validate(v)
            except KeyError:
                self.specific_params[k].validate(v)

        self._orig_kwargs.update(kw)
        self.kwargs = self._orig_kwargs.copy()

    def remove_params(self, *args):
        """
        Remove [possibly many] parameters from the track.

        E.g.,

        remove_params('color', 'visibility')
        """
        for a in args:
            self._orig_kwargs.pop(a)
        self.kwargs = self._orig_kwargs.copy()

    def add_subgroups(self, subgroups):
        """
        Update the subgroups for this track.

        :param subgroups:
            Dictionary of subgroups, e.g., {'celltype': 'K562', 'treatment':
                'a'}.  Each key must match a SubGroupDefinition name in the
                composite's subgroups list.  Each value must match a key in
                that SubGroupDefinition.mapping dictionary.
        """
        if subgroups is None:
            subgroups = {}
        assert isinstance(subgroups, dict)
        self.subgroups.update(subgroups)

    def __str__(self):
        s = []

        specific = []
        for name, parameter_obj in self.specific_params.items():
            value = self.kwargs.pop(name, None)
            if value is not None:
                if parameter_obj.validate(value):
                    specific.append("%s %s" % (name, value))

        for name, parameter_obj in self.params.items():
            value = self.kwargs.pop(name, None)
            if name == 'bigDataUrl':
                value = getattr(self, 'url', None)
            if value is not None:
                if parameter_obj.validate(value):
                    s.append("%s %s" % (name, value))

        s.extend(specific)

        # Handle subgroups differently depending on if this is a composite
        # track or not.
        s.extend(self._str_subgroups())

        if self.parent is not None:
            if isinstance(self.parent, BaseTrack):
                s.append('parent %s' % self.parent.name)

        if len(self.kwargs) > 0:
            raise ParameterError(
                "Unhandled keyword arguments: %s" % self.kwargs)

        self.kwargs = self._orig_kwargs.copy()
        return '\n'.join(s)

    def _render(self):
        if self._html:
            self._html.render()

    def _str_subgroups(self):
        """
        helper function to render subgroups as a string
        """
        if not self.subgroups:
            return ""
        return ['subGroups %s'
                % ' '.join(['%s=%s' % (k, v) for (k, v) in
                           self.subgroups.items()])]

    def validate(self):
        pass
        #assert self.tracktype in TRACKTYPES

    @property
    def html_fn(self):
        if self.remote_fn and self.trackdb:
            return os.path.join(
                os.path.dirname(self.trackdb.remote_fn),
                self.name + '.html')
        else:
            return None


class Track(BaseTrack):
    def __init__(self, url=None, *args, **kwargs):
        kwargs['bigDataUrl'] = kwargs.get('bigDataUrl', url)
        super(Track, self).__init__(*args, **kwargs)
        self._url = url

    @property
    def url(self):
        if self._url is not None:
            return self._url
        if self.remote_fn is None:
            return None
        return os.path.relpath(
            self.remote_fn,
            start=os.path.dirname(self.trackdb.remote_fn)
        )

    @url.setter
    def url(self, fn):
        self._url = fn


class CompositeTrack(BaseTrack):

    def __init__(self, *args, **kwargs):
        """
        Represents a composite track.  Subclasses :class:`Track`, and adds some
        extras.

        Add a view to this composite with :meth:`add_view`.

        Add a subtrack with :meth:`add_track`.

        Eventually, you'll need to make a :class:`trackdb.TrackDb` instance and
        add this composite to it with that instance's :meth:`add_tracks`
        method.

        If composite=True, then this track will be consider a composite
        parent for other tracks.  In this case, `subgroups` is a list of
        :class:`SubGroupDefinition` objects, each defining the possible
        values and display labels for the items in a group (for example,
        a celltype SubGroupDefinition would define the tags and titles for
        cell types).  In the string representation of this Track, subgroups
        end up looking like::

            subGroup1 view Views aln=Alignments sig=Signal
            subGroup2 celltype Cell_Type ES=embryonic k562=K562
        """
        super(CompositeTrack, self).__init__(*args, **kwargs)

        self.specific_params.update(constants.composite_track_fields)

        # TODO: are subtracks and views mutually exclusive, or can a composite
        # have both "view-ed" and "non-view-ed" subtracks?
        self.subtracks = []
        self.views = []

    def add_subgroups(self, subgroups):
        """
        Add a list of SubGroupDefinition objects to this composite.

        :param subgroups:
            List of SubGroupDefinition objects.
        """
        if subgroups is None:
            subgroups = {}
        _subgroups = {}
        for sg in subgroups:
            assert isinstance(sg, SubGroupDefinition)
            _subgroups[sg.name] = sg
        self.subgroups = _subgroups

    def add_subtrack(self, subtrack):
        """
        Add a child :class:`SubTrack`.
        """
        self.add_child(subtrack)
        self.subtracks.append(subtrack)

    def add_view(self, view):
        """
        Add a ViewTrack object to this composite.

        :param view:
            A ViewTrack object.
        """

        self.add_child(view)
        self.views.append(view)

    def _str_subgroups(self):
        """
        renders subgroups to a list of strings
        """
        s = []

        i = 0

        # if there are any views, there must be a subGroup1 view View tag=val
        # as the first one.  So create it automatically here
        if len(self.views) > 0:
            mapping = dict((i.view, i.view) for i in self.views)
            view_subgroup = SubGroupDefinition(
                name='view',
                label='Views',
                mapping=mapping)
            i += 1
            s.append('subGroup%s %s' % (i, view_subgroup))

        for subgroup in self.subgroups.values():
            i += 1
            s.append('subGroup%s %s' % (i, subgroup))
        return s

    def __str__(self):

        s = []

        s.append(super(CompositeTrack, self).__str__())
        s.append('compositeTrack on')

        for view in self.views:
            s.append("")
            for line in str(view).splitlines(False):
                s.append('    ' + line)

        for subtrack in self.subtracks:
            s.append("")
            for line in str(subtrack).splitlines(False):
                s.append('    ' + line)
        return "\n".join(s)


class ViewTrack(BaseTrack):
    def __init__(self, view, *args, **kwargs):
        """
        Represents a View track.  Subclasses :class:`Track`, and adds some
        extras.

        Upon being added to a CompositeTrack, the `view` tag will be checked.
        """
        self.view = view
        kwargs['view'] = view
        super(ViewTrack, self).__init__(*args, **kwargs)
        self.subtracks = []

    def add_subgroups(self, subgroups):
        if subgroups is None:
            subgroups = {}
        else:
            raise ValueError('not sure if Views can have subgroups?')

    def add_tracks(self, subtracks):
        """
        Add tracks to this view.

        :param subtracks:
            A single Track instance or an iterable of them.
        """
        if isinstance(subtracks, Track):
            subtracks = [subtracks]
        for subtrack in subtracks:
            subtrack.subgroups['view'] = self.view
            self.add_child(subtrack)
            self.subtracks.append(subtrack)

    def _str_subgroups(self):
        return ""

    def __str__(self):
        s = []
        view_specific = []
        for name, parameter_obj in constants.view_track_fields.items():
            value = self.kwargs.pop(name, None)
            if value is not None:
                if parameter_obj.validate(value):
                    view_specific.append("%s %s" % (name, value))

        s.append(super(ViewTrack, self).__str__())
        s.extend(view_specific)

        for subtrack in self.subtracks:
            s.append("")
            for line in str(subtrack).splitlines(False):
                s.append('        ' + line)
        return '\n'.join(s)


class SuperTrack(BaseTrack):
    def __init__(self, *args, **kwargs):
        """
        Represents a Super track. Subclasses :class:`Track`, and adds some
        extras.

        Super tracks are container tracks (Folders) that group tracks. They are
        used to control visualization of a set of related data.

        Eventually, you'll need to make a :class:`trackdb.TrackDb` instance and
        add this supertrack to it with that instance's :meth:`add_tracks`
        method.
        """
        super(SuperTrack, self).__init__(*args, **kwargs)
        self.subtracks = []

    def add_track(self, subtrack):
        """
        Add a child :class:`SubTrack` to this supertrack.
        """
        self.add_child(subtrack)
        self.subtracks.append(subtrack)

    def __str__(self):

        s = []

        s.append(super(SuperTrack, self).__str__())
        s.append('superTrack on')

        # Removed subtracks for Supertrack because composite tracks can be
        # within the supertrack.  This is also the recommendation from UCSC
        for subtrack in self.subtracks:
            s.append("")
            for line in str(subtrack).splitlines(False):
                s.append(line)
        return '\n'.join(s)


class AggregateTrack(BaseTrack):
    def __init__(self, aggregate, *args, **kwargs):
        """
        Represents an Aggregate or Overlay track. Subclasses :class:`Track`,
        adds some extras.

        Aggregate tracks allow closley related tracks to be viewed as a single
        track.

        Eventually, you'll need to make a :class:`trackdb.TrackDb` instance and
        add this supertrack to it with that instance's :meth:`add_tracks`
        method.
        """

        self.aggregate = aggregate
        kwargs['aggregate'] = aggregate
        super(AggregateTrack, self).__init__(*args, **kwargs)
        self.specific_params.update(constants.aggregate_track_fields)
        self.subtracks = []

    def add_subtrack(self, subtrack):
        """
        Add a child :class:`SubTrack` to this aggregrate.
        """
        self.add_child(subtrack)
        self.subtracks.append(subtrack)

    def __str__(self):

        s = []

        s.append(super(AggregateTrack, self).__str__())
        s.append('container multiWig')

        for subtrack in self.subtracks:
            s.append("")
            for line in str(subtrack).splitlines(False):
                s.append('    ' + line)
        return "\n".join(s)


class HTMLDoc(HubComponent):
    def __init__(self, contents):
        """
        Represents an HTML file used for documentation.

        Handles local/remote/url filenames when connected to a Track and
        CompositeTrack
        """
        self.contents = contents
        self._local_fn = None
        self._remote_fn = None
        super(HTMLDoc, self).__init__()

    def validate(self):
        return

    @property
    def local_fn(self):
        if (self.trackdb is None) or (self.track is None):
            return None
        return os.path.join(
            os.path.dirname(self.trackdb.local_fn),
            self.track.name + '.html')

    @property
    def remote_fn(self):
        if (self.trackdb is None) or (self.track is None):
            return None
        return os.path.join(
            os.path.dirname(self.trackdb.remote_fn),
            self.track.name + '.html')

    @property
    def trackdb(self):
        obj, level = self.root(cls=trackdb.TrackDb)
        return obj

    @property
    def track(self):
        obj, level = self.root(cls=BaseTrack)
        return obj

    def _render(self):
        if not self.trackdb:
            raise ValueError("HTMLDoc object must be connected to a "
                             "BaseTrack subclass instance and a TrackDb "
                             "instance")
        dirname = os.path.dirname(self.local_fn)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        fout = open(self.local_fn, 'w')
        fout.write(str(self))
        fout.close()
        return fout.name

    def __str__(self):
        return self.contents
